import polars as pl

from src.recipies.constants import Backend
from src.recipies.recipe import Recipe


def test_empty_prep_return_df(example_pl_df):
    rec = Recipe(example_pl_df)
    assert isinstance(rec.prep(), pl.DataFrame)


def test_empty_bake_return_df(example_pl_df):
    rec = Recipe(example_pl_df)
    assert isinstance(rec.bake(), pl.DataFrame)


def test_prep_bake_same_result(example_pl_df, example_pl_recipe):
    example2 = example_pl_df.clone()
    output1 = example_pl_recipe.prep(example_pl_df)
    output2 = example_pl_recipe.bake(example2)
    assert output1.equals(output2)


def test_init_roles(example_pl_df):
    rec = Recipe(example_pl_df, ["y"], ["x1", "x2", "x3"], ["id"], ["time"])
    assert rec.data.roles["y"] == ["outcome"]
    assert rec.data.roles["x1"] == ["predictor"]
    assert rec.data.roles["x2"] == ["predictor"]
    assert rec.data.roles["x3"] == ["predictor"]
    assert rec.data.roles["time"] == ["sequence"]
    assert rec.data.roles["id"] == ["group"]


def test_inferring_backend_recipe(example_df):
    ing = Recipe(example_df)
    if isinstance(example_df, pl.DataFrame):
        assert ing.get_backend() == Backend.POLARS
    else:
        assert ing.get_backend() == Backend.PANDAS


def test_explicit_backend_recipe(example_df):
    ing = Recipe(example_df, backend=Backend.PANDAS)
    assert ing.get_backend() == Backend.PANDAS
    ing = Recipe(example_df, backend=Backend.POLARS)
    assert ing.get_backend() == Backend.POLARS


def test_backend_ingredients_recipe(example_ingredients):
    rec = Recipe(example_ingredients)
    assert rec.get_backend() == example_ingredients.get_backend()



import pytest
import numpy as np
import polars as pl
from src.recipies.recipe import Recipe
from src.recipies.ingredients import Ingredients
import pandas as pd
from src.recipies.constants import Backend


@pl.api.register_dataframe_namespace("pd")
class PolarsPd:
    def __init__(self, df: pl.DataFrame):
        self.df = df.to_pandas()


@pytest.fixture
def example_pl_df():
    rand_state = np.random.RandomState(42)
    timecolumn = pl.from_pandas(pd.to_timedelta(np.concatenate((np.arange(6), np.arange(4))), unit="h"))
    # timecolumn = pl.concat([pl.datetime_range(datetime(MINYEAR, 1, 1,0), datetime(MINYEAR, 1, 1,5), "1h", eager=True),
    #           pl.datetime_range(datetime(MINYEAR, 1, 1,0), datetime(MINYEAR, 1, 1,3), "1h", eager=True)])

    df = pl.DataFrame(
        {
            "id": [1] * 6 + [2] * 4,
            "time": timecolumn,
            "y": rand_state.normal(size=(10,)),
            "x1": rand_state.normal(loc=10, scale=5, size=(10,)),
            "x2": rand_state.binomial(n=1, p=0.3, size=(10,)),
            "x3": pl.Series(["a", "b", "c", "a", "c", "b", "c", "a", "b", "c"], dtype=pl.Categorical),
            "x4": pl.Series(["x", "y", "y", "x", "y", "y", "x", "x", "y", "x"], dtype=pl.Categorical),
        }
    )
    return df


@pytest.fixture
def example_pd_df():
    rand_state = np.random.RandomState(42)
    df = pd.DataFrame(
        {
            "id": [1] * 6 + [2] * 4,
            "time": pd.to_timedelta(np.concatenate((np.arange(6), np.arange(4))), unit="h"),
            "y": rand_state.normal(size=(10,)),
            "x1": rand_state.normal(loc=10, scale=5, size=(10,)),
            "x2": rand_state.binomial(n=1, p=0.3, size=(10,)),
            "x3": pd.Series(["a", "b", "c", "a", "c", "b", "c", "a", "b", "c"], dtype="category"),
            "x4": pd.Series(["x", "y", "y", "x", "y", "y", "x", "x", "y", "x"], dtype="category"),
        }
    )
    return df


@pytest.fixture(params=["example_pl_df", "example_pd_df"])
def example_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def example_pl_ingredients(example_pl_df):
    return Ingredients(example_pl_df)


@pytest.fixture
def example_pd_ingredients(example_pd_df):
    return Ingredients(example_pd_df, backend=Backend.PANDAS)


@pytest.fixture(params=["example_pl_ingredients", "example_pd_ingredients"])
def example_ingredients(request):
    return request.getfixturevalue(request.param)


# @pytest.fixture
# def example_ingredients(example_pl_df):
#     return Ingredients(example_pl_df, backend=Backend.PANDAS)


@pytest.fixture()
def example_pl_recipe(example_pl_df):
    return Recipe(example_pl_df, ["y"], ["x1", "x2", "x3", "x4"], ["id"], ["time"])


@pytest.fixture()
def example_pd_recipe(example_pd_df):
    return Recipe(example_pd_df, ["y"], ["x1", "x2", "x3", "x4"], ["id"], ["time"])


@pytest.fixture()
def example_recipe_w_nan(example_df):
    example_df.loc[[1, 2, 4, 7], "x1"] = np.nan
    return Recipe(example_df, ["y"], ["x1", "x2", "x3", "x4"], ["id"], ["time"])



import pytest
from src.recipies.constants import Backend

from src.recipies.selector import (
    Selector,
    all_outcomes,
    all_of,
    regex_names,
    select_groups,
    select_sequence,
    starts_with,
    ends_with,
    contains,
    has_role,
    has_type,
    all_numeric_predictors,
    all_predictors,
    intersection,
    enlist_str,
)


def test_no_description():
    with pytest.raises(TypeError) as e_info:
        Selector()
    assert e_info.match("missing 1 required positional argument")


def test_not_ingredients(example_pl_df):
    with pytest.raises(TypeError) as e_info:
        Selector("test step")(example_pl_df)
    assert e_info.match("Expected Ingredients")


def test_intersection():
    assert intersection(["a", "b"], ["b", "c"]) == ["b"]


def test_enlist_str():
    assert enlist_str("string") == ["string"]


def test_enlist_str_list():
    assert enlist_str(["string1", "string2"]) == ["string1", "string2"]


def test_enlist_str_None():
    assert enlist_str(None) is None


def test_enlist_str_other():
    with pytest.raises(TypeError) as e_info:
        enlist_str({"k": "string"})
    assert e_info.match("Expected str or list of str")


def test_enlist_str_other_list():
    with pytest.raises(TypeError) as e_info:
        enlist_str(["outer", {"k": "inner"}])
    assert e_info.match("Only lists of str are allowed.")


def test_all_of(example_ingredients):
    sel = all_of(["y", "x1"])
    assert sel(example_ingredients) == ["y", "x1"]


def test_regex_names(example_ingredients):
    sel = regex_names(r"^x\d")
    assert sel(example_ingredients) == ["x1", "x2", "x3", "x4"]


def test_starts_with(example_ingredients):
    sel = starts_with("x")
    assert sel(example_ingredients) == ["x1", "x2", "x3", "x4"]


def test_ends_with(example_ingredients):
    sel = ends_with("1")
    assert sel(example_ingredients) == ["x1"]


def test_contains(example_ingredients):
    sel = contains("i")
    assert sel(example_ingredients) == ["id", "time"]


def test_has_role(example_ingredients):
    example_ingredients.update_role("x1", "predictor")
    example_ingredients.update_role("x2", "predictor")
    sel = has_role("predictor")
    assert sel(example_ingredients) == ["x1", "x2"]


def test_has_type(example_ingredients):
    # sel = has_type("Float64")
    # sel = has_type(pl.Float64)
    if example_ingredients.get_backend() == Backend.POLARS:
        sel = has_type("Float64")
    else:
        sel = has_type("float64")
    assert sel(example_ingredients) == ["y", "x1"]


# def test_has_type_pl(example_ingredients):
#     sel = has_type(pl.Float64)
#     assert sel(example_ingredients) == ["y", "x1"]
def test_all_predictors(example_ingredients):
    example_ingredients.update_role("x1", "predictor")
    example_ingredients.update_role("x2", "predictor")
    sel = all_predictors()
    assert sel(example_ingredients) == ["x1", "x2"]


def test_all_numeric_predictors(example_ingredients):
    example_ingredients.update_role("x1", "predictor")
    example_ingredients.update_role("x2", "predictor")
    sel = all_numeric_predictors(backend=example_ingredients.get_backend())
    assert sel(example_ingredients) == ["x1", "x2"]


def test_all_outcomes(example_ingredients):
    example_ingredients.update_role("y", "outcome")
    sel = all_outcomes()
    assert sel(example_ingredients) == ["y"]


def test_select_groups(example_ingredients):
    example_ingredients.update_role("id", "group")
    assert select_groups(example_ingredients) == ["id"]


def test_select_sequence(example_ingredients):
    example_ingredients.update_role("time", "sequence")
    assert select_sequence(example_ingredients) == ["time"]



from datetime import datetime, MINYEAR

import pandas as pd
import pytest
import polars as pl
import numpy as np
from sklearn.preprocessing import (
    Binarizer,
    FunctionTransformer,
    KBinsDiscretizer,
    LabelBinarizer,
    LabelEncoder,
    MaxAbsScaler,
    MinMaxScaler,
    Normalizer,
    OneHotEncoder,
    OrdinalEncoder,
    PolynomialFeatures,
    PowerTransformer,
    QuantileTransformer,
    RobustScaler,
    StandardScaler,
    SplineTransformer,
)
from sklearn.experimental import enable_iterative_imputer  # noqa: F401
from sklearn.impute import SimpleImputer, KNNImputer, IterativeImputer, MissingIndicator

from src.recipies.ingredients import Ingredients
from src.recipies.recipe import Recipe
from src.recipies.selector import all_numeric_predictors, has_type, has_role, all_of
from src.recipies.step import (
    StepSklearn,
    StepHistorical,
    Accumulator,
    StepImputeFill,
    StepScale,
    StepResampling,
    StepImputeFastZeroFill,
    StepImputeFastForwardFill,
)
from src.recipies.constants import Backend


@pytest.fixture()
def example_recipe(example_ingredients):
    return Recipe(example_ingredients, ["y"], ["x1", "x2", "x3", "x4"], ["id"], ["time"])  # FIXME: add squence when merged


@pytest.fixture()
def example_recipe_w_nan(example_ingredients):
    example_ingredients[[2, 4, 6], "x2"] = np.nan
    return Recipe(example_ingredients, ["y"], ["x1", "x2", "x3", "x4"], ["id"], ["time"])  # FIXME: add squence when merged


def test_no_group_for_group_step(example_ingredients):
    rec = Recipe(example_ingredients, ["y"], ["x1", "x2"])
    rec.add_step(StepImputeFill(value=0))
    rec.prep()


class TestStepResampling:
    def test_step_grouped(self, example_df):
        # Using group role
        pre_sampling_len = example_df.shape[0]
        if isinstance(example_df, pl.DataFrame):
            backend = Backend.POLARS
            timecolumn = pl.concat(
                [
                    pl.datetime_range(datetime(MINYEAR, 1, 1, 0), datetime(MINYEAR, 1, 1, 5), "1h", eager=True),
                    pl.datetime_range(datetime(MINYEAR, 1, 1, 0), datetime(MINYEAR, 1, 1, 3), "1h", eager=True),
                ]
            )
            example_df = example_df.with_columns(time=timecolumn)
        else:
            backend = Backend.PANDAS
        rec = Recipe(example_df, ["y"], ["x1", "x2"], ["id"], ["time"], backend=backend)
        resampling_dict = {all_numeric_predictors(backend): Accumulator.MEAN}
        rec.add_step(StepResampling("2h", accumulator_dict=resampling_dict))
        df = rec.bake()
        assert df.shape[0] == pre_sampling_len / 2

    def test_step_wo_selectors(self, example_df):
        # Using group role and without supplying any selectors
        pre_sampling_len = example_df.shape[0]
        if isinstance(example_df, pl.DataFrame):
            backend = Backend.POLARS
            timecolumn = pl.concat(
                [
                    pl.datetime_range(datetime(MINYEAR, 1, 1, 0), datetime(MINYEAR, 1, 1, 5), "1h", eager=True),
                    pl.datetime_range(datetime(MINYEAR, 1, 1, 0), datetime(MINYEAR, 1, 1, 3), "1h", eager=True),
                ]
            )
            example_df = example_df.with_columns(time=timecolumn)
        else:
            backend = Backend.PANDAS
        rec = Recipe(example_df, ["y"], ["x1", "x2"], ["id"], ["time"], backend=backend)
        rec.add_step(StepResampling("2h"))
        df = rec.bake()
        assert df.shape[0] == pre_sampling_len / 2

    # Todo: check if desired behaviour to have no group role
    def test_step_ungrouped(self, example_df):
        # Without using group role
        if isinstance(example_df, pl.DataFrame):
            timecolumn = pl.concat(
                [
                    pl.datetime_range(datetime(MINYEAR, 1, 1, 0), datetime(MINYEAR, 1, 1, 5), "1h", eager=True),
                    pl.datetime_range(datetime(MINYEAR, 1, 1, 0), datetime(MINYEAR, 1, 1, 3), "1h", eager=True),
                ]
            )
            example_df = example_df.with_columns(time=timecolumn)
            example_df = example_df.drop("id")
            example_df = example_df.unique(subset="time")
        else:
            # Pandas
            example_df.drop("id", axis=1, inplace=True)
            example_df = example_df.drop_duplicates(subset="time", inplace=False, keep="first")
        pre_sampling_len = example_df.shape[0]
        rec = Recipe(example_df, ["y"], ["x1", "x2"])
        rec.update_roles("time", "sequence")
        resampling_dict = {all_numeric_predictors(): Accumulator.MEAN}
        rec.add_step(StepResampling("2h", accumulator_dict=resampling_dict))
        df = rec.bake()
        assert df.shape[0] == (pre_sampling_len / 2)


class TestStepHistorical:
    def test_step(self, example_df):
        rec = Recipe(Ingredients(example_df), ["y"], ["x1", "x2"], ["id"])
        rec.add_step(StepHistorical(sel=all_of(["x1", "x2"]), fun=Accumulator.MIN, suffix="min"))
        rec.add_step(StepHistorical(sel=all_of(["x1", "x2"]), fun=Accumulator.MAX, suffix="max"))
        rec.add_step(StepHistorical(sel=all_of(["x1", "x2"]), fun=Accumulator.MEAN, suffix="mean"))
        rec.add_step(StepHistorical(sel=all_of(["x1", "x2"]), fun=Accumulator.MEDIAN, suffix="median"))
        rec.add_step(StepHistorical(sel=all_of(["x1", "x2"]), fun=Accumulator.COUNT, suffix="count"))
        rec.add_step(StepHistorical(sel=all_of(["x1", "x2"]), fun=Accumulator.VAR, suffix="var"))
        df = rec.bake()
        if rec.get_backend() == Backend.POLARS:
            assert df["x1_min"][-1] == df.filter(pl.col("id") == 2).select(pl.col("x1")).min().item()
            assert df["x1_max"][-1] == df.filter(pl.col("id") == 2).select(pl.col("x1")).max().item()
            # assert df["x1_mean"][-1] == df.filter(pl.col("id") == 2).select(pl.col("x1")).mean().item()
            # With approximate equality
            assert df["x1_mean"][-1] == pytest.approx(df.filter(pl.col("id") == 2).select(pl.col("x1")).mean().item())
            assert df["x1_median"][-1] == df.filter(pl.col("id") == 2).select(pl.col("x1")).median().item()
            assert df["x1_count"][-1] == df.filter(pl.col("id") == 2).select(pl.col("x1")).count().item()
            # somehow we get a rounding difference between these two values
            assert (
                df["x1_var"].round(2)[-1]
                == df.filter(pl.col("id") == 2).select(pl.col("x1")).var().to_series().round(2).item()
            )
        else:
            assert df["x1_min"].iloc[-1] == df["x1"].loc[df["id"] == 2].min()
            assert df["x2_max"].iloc[-1] == df["x2"].loc[df["id"] == 2].max()
            assert df["x2_mean"].iloc[-1] == df["x2"].loc[df["id"] == 2].mean()
            assert df["x1_median"].iloc[-1] == df["x1"].loc[df["id"] == 2].median()
            assert df["x1_count"].iloc[-1] == df["x1"].loc[df["id"] == 2].count()
            assert df["x2_var"].iloc[-1] == df["x2"].loc[df["id"] == 2].var()


class TestImputeSteps:
    def test_impute_fill(self, example_recipe_w_nan):
        example_recipe_w_nan.add_step(StepImputeFill(strategy="forward"))
        backend = example_recipe_w_nan.get_backend()
        res = example_recipe_w_nan.prep()
        nan_list = [0, 1, 1, 0, 0, 0, np.nan, 0, 0, 1]
        exp = (
            pl.Series("x2", nan_list, dtype=pl.Int32, strict=False)
            if backend == Backend.POLARS
            else pd.Series(nan_list, dtype="float64")
        )
        assert res["x2"].equals(exp)
        example_recipe_w_nan.add_step(StepImputeFill(sel=all_numeric_predictors(backend), value=0))
        res = example_recipe_w_nan.prep()
        imputed_list = [0, 1, 1, 0, 0, 0, 0, 0, 0, 1]
        exp = (
            pl.Series("x2", imputed_list, pl.Int32, strict=False)
            if backend == Backend.POLARS
            else pd.Series(imputed_list, dtype="float64")
        )
        assert res["x2"].equals(exp)

    def test_fast_zero_fill(self, example_recipe_w_nan):
        backend = example_recipe_w_nan.get_backend()
        example_recipe_w_nan.add_step(StepImputeFastZeroFill(sel=all_numeric_predictors(backend)))
        if backend == Backend.POLARS:
            with pytest.raises(ValueError) as e_info:
                res = example_recipe_w_nan.prep()
            assert e_info.match("Backend.POLARS not supported by this step.")
        else:
            res = example_recipe_w_nan.prep()
            imputed_list = [0, 1, 0, 0, 0, 0, 0, 0, 0, 1]
            exp = pd.Series(imputed_list, dtype="float64")
            assert res["x2"].equals(exp)

    def test_fast_forward_fill(self, example_recipe_w_nan):
        backend = example_recipe_w_nan.get_backend()
        example_recipe_w_nan.add_step(StepImputeFastForwardFill(sel=all_numeric_predictors(backend)))
        if backend == Backend.POLARS:
            with pytest.raises(ValueError) as e_info:
                res = example_recipe_w_nan.prep()
            assert e_info.match("Backend.POLARS not supported by this step.")
        else:
            res = example_recipe_w_nan.prep()
            imputed_list = [0, 1, 1, 0, 0, 0, np.nan, 0, 0, 1]
            exp = pd.Series(imputed_list, dtype="float64")
            assert res["x2"].equals(exp)


class TestScaleStep:
    def test_scale_step_default(self, example_recipe_w_nan):
        example_recipe_w_nan.add_step(StepScale(all_numeric_predictors(backend=example_recipe_w_nan.get_backend())))
        res = example_recipe_w_nan.prep()
        assert abs(res["x1"].mean()) < 0.00001
        assert abs(res["x2"].mean()) < 0.00001

    def test_scale_step_w_args(self, example_recipe):
        example_recipe.add_step(
            StepScale(all_numeric_predictors(backend=example_recipe.get_backend()), with_mean=False, with_std=False)
        )
        res = example_recipe.prep()
        assert abs(res["x1"].mean()) > 1
        assert abs(res["x1"].var()) > 1.5

    def test_scale_step_w_sel(self, example_recipe):
        example_recipe.add_step(StepScale(sel=all_of(["x2"])))
        res = example_recipe.prep()
        assert abs(res["x2"].mean()) < 0.00001
        assert abs(res["x1"].mean()) > 1


class TestSklearnStep:
    @pytest.fixture()
    def example_recipe_w_categorical_label(self, example_df):
        if isinstance(example_df, pl.DataFrame):
            example_df = example_df.with_columns(
                y=pl.Series(["a", "b", "c", "a", "c", "b", "c", "a", "b", "c"], dtype=pl.Categorical)
            )
        else:
            example_df["y"] = pd.Categorical(["a", "b", "c", "a", "c", "b", "c", "a", "b", "c"])
        return Recipe(Ingredients(example_df), ["y"], ["x1", "x2", "x3", "x4"], ["id"])  # FIXME: add squence when merged

    def test_simple_imputer(self, example_recipe_w_nan):
        backend = example_recipe_w_nan.get_backend()
        example_recipe_w_nan.add_step(StepSklearn(SimpleImputer(strategy="constant", fill_value=0)))
        df = example_recipe_w_nan.prep()
        assert (
            (df[[2, 4, 6], "x2"].to_numpy() == np.full(3, 0)).all()
            if backend == backend.POLARS
            else (df.loc[[2, 4, 6], "x2"] == 0).all()
        )

    def test_knn_imputer(self, example_recipe_w_nan):
        backend = example_recipe_w_nan.get_backend()
        example_recipe_w_nan.add_step(StepSklearn(KNNImputer(), sel=all_numeric_predictors(backend)))
        df = example_recipe_w_nan.prep()
        assert (
            (~np.isnan(df[[2, 4, 6], "x2"].to_numpy())).all()
            if backend == backend.POLARS
            else (~np.isnan(df.loc[[2, 4, 6], "x2"])).all()
        )

    def test_iterative_imputer(self, example_recipe_w_nan):
        backend = example_recipe_w_nan.get_backend()
        example_recipe_w_nan.add_step(StepSklearn(IterativeImputer(), sel=all_numeric_predictors(backend)))
        df = example_recipe_w_nan.prep()
        assert (
            (~np.isnan(df[[2, 4, 6], "x2"].to_numpy())).all()
            if backend == backend.POLARS
            else (~np.isnan(df.loc[[2, 4, 6], "x2"])).all()
        )

    def test_missing_indicator(self, example_recipe_w_nan):
        backend = example_recipe_w_nan.get_backend()
        example_recipe_w_nan.add_step(
            StepSklearn(MissingIndicator(features="all"), sel=all_numeric_predictors(backend), in_place=False)
        )
        df = example_recipe_w_nan.prep()
        assert (
            (df[[2, 4, 6], "MissingIndicator_x2"].to_numpy()).all()
            if backend == backend.POLARS
            else (df.loc[[2, 4, 6], "MissingIndicator_x2"]).all()
        )

    def test_standard_scaler(self, example_recipe):
        backend = example_recipe.get_backend()
        example_recipe.add_step(StepSklearn(StandardScaler(), sel=all_numeric_predictors(backend)))
        df = example_recipe.prep()
        assert abs(df["x1"].mean()) < 0.00001
        assert abs(df["x2"].mean()) < 0.00001

    def test_min_max_scaler(self, example_recipe):
        backend = example_recipe.get_backend()
        example_recipe.add_step(StepSklearn(MinMaxScaler(), sel=all_numeric_predictors(backend)))
        df = example_recipe.prep()
        assert ((0 <= df["x1"]) & (df["x1"] <= 1)).all()
        assert ((0 <= df["x2"]) & (df["x2"] <= 1)).all()

    def test_max_abs_scaler(self, example_recipe):
        backend = example_recipe.get_backend()
        example_recipe.add_step(StepSklearn(MaxAbsScaler(), sel=all_numeric_predictors(backend)))
        df = example_recipe.prep()
        assert ((-1 <= df["x1"]) & (df["x1"] <= 1)).all()
        assert ((-1 <= df["x2"]) & (df["x2"] <= 1)).all()

    def test_robust_scaler(self, example_recipe):
        backend = example_recipe.get_backend()
        example_recipe.add_step(StepSklearn(RobustScaler(), sel=all_numeric_predictors(backend)))
        df = example_recipe.prep()
        assert df["x1"].median() < 10e-12
        assert df["x2"].median() < 10e-12

    def test_binarizer(self, example_recipe):
        backend = example_recipe.get_backend()
        example_recipe.add_step(StepSklearn(Binarizer(), sel=all_numeric_predictors(backend=backend)))
        df = example_recipe.prep()
        assert (df["x1"].is_in([0, 1])).all() if backend == Backend.POLARS else (df["x1"].isin([0, 1])).all()
        assert (df["x2"].is_in([0, 1])).all() if backend == Backend.POLARS else (df["x2"].isin([0, 1])).all()

    def test_normalizer(self, example_recipe):
        example_recipe.add_step(StepSklearn(Normalizer(), sel=all_numeric_predictors(backend=example_recipe.get_backend())))
        df = example_recipe.prep()
        assert ((0 <= df["x1"]) & (df["x1"] <= 1)).all()
        assert ((0 <= df["x2"]) & (df["x2"] <= 1)).all()

    def test_k_bins_binarizer(self, example_recipe):
        backend = example_recipe.get_backend()
        example_recipe.add_step(
            StepSklearn(
                KBinsDiscretizer(n_bins=2, strategy="uniform", encode="ordinal"),
                sel=all_numeric_predictors(backend=example_recipe.get_backend()),
                in_place=False,
            )
        )
        df = example_recipe.prep()
        assert (
            (df["KBinsDiscretizer_x1"].is_in([0, 1])).all()
            if backend == Backend.POLARS
            else (df["KBinsDiscretizer_x1"].isin([0, 1])).all()
        )
        assert (
            (df["KBinsDiscretizer_x2"].is_in([0, 1])).all()
            if backend == Backend.POLARS
            else (df["KBinsDiscretizer_x2"].isin([0, 1])).all()
        )

    def test_quantile_transformer(self, example_recipe):
        example_recipe.add_step(
            StepSklearn(QuantileTransformer(n_quantiles=10), sel=all_numeric_predictors(example_recipe.get_backend()))
        )
        df = example_recipe.prep()
        assert ((0 <= df["x1"]) & (df["x1"] <= 1)).all()
        assert ((0 <= df["x2"]) & (df["x2"] <= 1)).all()

    def test_ordinal_encoder(self, example_recipe):
        backend = example_recipe.get_backend()
        example_recipe.add_step(
            StepSklearn(
                OrdinalEncoder(),
                sel=has_type([str(pl.Categorical(ordering="physical")) if backend == backend.POLARS else "category"]),
                in_place=False,
            )
        )
        df = example_recipe.prep()
        # FIXME assert correct number of new columns
        assert ((0 <= df["OrdinalEncoder_x3"]) & (df["OrdinalEncoder_x4"] <= 2)).all()

    def test_onehot_encoder(self, example_recipe):
        backend = example_recipe.get_backend()
        example_recipe.add_step(
            StepSklearn(
                OneHotEncoder(sparse_output=False),
                sel=has_type([str(pl.Categorical(ordering="physical")) if backend == backend.POLARS else "category"]),
                in_place=False,
            )
        )
        df = example_recipe.prep()
        if backend == backend.POLARS:
            assert (df["OneHotEncoder_1"].is_in([0, 1])).all()
            assert (df["OneHotEncoder_2"].is_in([0, 1])).all()
            assert (df["OneHotEncoder_3"].is_in([0, 1])).all()
            assert (df["OneHotEncoder_4"].is_in([0, 1])).all()
            assert (df["OneHotEncoder_5"].is_in([0, 1])).all()
        else:
            assert (df["OneHotEncoder_1"].isin([0, 1])).all()
            assert (df["OneHotEncoder_2"].isin([0, 1])).all()
            assert (df["OneHotEncoder_3"].isin([0, 1])).all()
            assert (df["OneHotEncoder_4"].isin([0, 1])).all()
            assert (df["OneHotEncoder_5"].isin([0, 1])).all()

    def test_label_encoder(self, example_recipe_w_categorical_label):
        example_recipe_w_categorical_label.add_step(StepSklearn(LabelEncoder(), sel=has_role(["outcome"]), columnwise=True))
        df = example_recipe_w_categorical_label.prep()
        assert ((0 <= df["y"]) & (df["y"] <= 2)).all()

    def test_label_binarizer(self, example_recipe_w_categorical_label):
        example_recipe_w_categorical_label.add_step(
            StepSklearn(LabelBinarizer(), sel=has_role(["outcome"]), columnwise=True, in_place=False, role="outcome")
        )
        df = example_recipe_w_categorical_label.prep()
        if example_recipe_w_categorical_label.get_backend() == Backend.POLARS:
            assert (df["LabelBinarizer_y_1"].is_in([0, 1])).all()
            assert (df["LabelBinarizer_y_2"].is_in([0, 1])).all()
            assert (df["LabelBinarizer_y_3"].is_in([0, 1])).all()
        else:
            assert (df["LabelBinarizer_y_1"].isin([0, 1])).all()
            assert (df["LabelBinarizer_y_2"].isin([0, 1])).all()
            assert (df["LabelBinarizer_y_3"].isin([0, 1])).all()

    def test_spline_transformer(self, example_recipe):
        backend = example_recipe.get_backend()
        example_recipe.add_step(StepSklearn(SplineTransformer(), sel=all_numeric_predictors(backend), in_place=False))
        df = example_recipe.prep()
        # FIXME assert correct number of new columns
        assert not df["SplineTransformer_1"].is_empty() if backend == Backend.POLARS else not df["SplineTransformer_1"].empty

    def test_polynomial_features(self, example_recipe):
        backend = example_recipe.get_backend()
        example_recipe.add_step(StepSklearn(PolynomialFeatures(), sel=all_numeric_predictors(backend), in_place=False))
        df = example_recipe.prep()
        # FIXME assert correct number of new columns
        assert not df["PolynomialFeatures_1"].is_empty() if backend == Backend.POLARS else not df["PolynomialFeatures_1"].empty

    def test_power_transformer(self, example_recipe):
        backend = example_recipe.get_backend()
        example_recipe.add_step(StepSklearn(PowerTransformer(), sel=all_numeric_predictors(backend), in_place=False))
        df = example_recipe.prep()
        # FIXME assert correct number of new columns
        assert not df["PowerTransformer_x1"].is_empty() if backend == Backend.POLARS else not df["PowerTransformer_x1"].empty

    def test_function_transformer(self, example_recipe):
        backend = example_recipe.get_backend()
        example_recipe.add_step(
            StepSklearn(
                FunctionTransformer(np.log1p), sel=all_numeric_predictors(example_recipe.get_backend()), in_place=False
            )
        )
        df = example_recipe.prep()
        # FIXME assert correct number of new columns
        assert (
            not df["FunctionTransformer_x1"].is_empty()
            if backend == Backend.POLARS
            else not df["FunctionTransformer_x1"].empty
        )

    def test_wrong_columnwise(self, example_df):
        if isinstance(example_df, pl.DataFrame):
            example_df = example_df.with_columns(
                y=pl.Series(["a", "b", "c", "a", "c", "b", "c", "a", "b", "c"], dtype=pl.Categorical)
            )
            example_df = example_df.with_columns(
                y1=pl.Series(["a", "b", "c", "a", "c", "b", "c", "a", "b", "c"], dtype=pl.Categorical)
            )
        else:
            example_df["y"] = pd.Categorical(["a", "b", "c", "a", "c", "b", "c", "a", "b", "c"])
            example_df["y1"] = pd.Categorical(["a", "b", "c", "a", "c", "b", "c", "a", "b", "c"])
        rec = Recipe(Ingredients(example_df), ["y", "y1"], ["x1", "x2", "x3"], ["id"], ["time"])
        rec.add_step(StepSklearn(LabelEncoder(), sel=has_role(["outcome"]), columnwise=False))
        with pytest.raises(ValueError) as exc_info:
            rec.prep()
        assert "columnwise=True" in str(exc_info.value)

    def test_wrong_in_place(self, example_recipe):
        backend = example_recipe.get_backend()
        example_recipe.add_step(
            StepSklearn(
                OneHotEncoder(sparse_output=False),
                sel=has_type([str(pl.Categorical(ordering="physical")) if backend == backend.POLARS else "category"]),
                in_place=True,
            )
        )
        with pytest.raises(ValueError) as exc_info:
            example_recipe.prep()
        assert "in_place=False" in str(exc_info.value)

    def test_sparse_output_error(self, example_recipe):
        backend = example_recipe.get_backend()
        example_recipe.add_step(
            StepSklearn(
                OneHotEncoder(sparse_output=True),
                sel=has_type([str(pl.Categorical(ordering="physical")) if backend == backend.POLARS else "category"]),
                in_place=False,
            )
        )
        with pytest.raises(TypeError) as exc_info:
            example_recipe.prep()
        assert "sparse_output=False" in str(exc_info.value)



from __future__ import annotations
from collections import Counter
from copy import copy
from itertools import chain
from typing import Union

import pandas as pd
import polars as pl

from .ingredients import Ingredients
from .selector import select_groups
from .step import Step
from .constants import Backend


class Recipe:
    """Recipe for preprocessing data

    A Recipe object combines a pandas-like Ingredients object with one or more
    sklearn-inspired transformation Steps to turn into a model-ready input.

    Args:
        data: data to be preprocessed.
        outcomes: names of columns in data that are assigned the 'outcome' role
        predictors: names of columns in data that should be assigned the 'predictor' role
        groups: names of columns in data that should be assigned the 'group' role
        sequence: names of columns in data that should be assigned the 'sequence' role
    """

    columns = None
    roles = None

    def __init__(
        self,
        data: Ingredients | pl.DataFrame | pd.DataFrame,
        outcomes: Union[str, list[str]] = None,
        predictors: Union[str, list[str]] = None,
        groups: Union[str, list[str]] = None,
        sequences: Union[str, list[str]] = None,
        backend: Backend = None,
    ):
        if not isinstance(data, Ingredients):
            try:
                data = Ingredients(data, backend=backend)
            except Exception as e:
                raise (f"Expected Ingredients, got {data.__class__} {e}")
        self.data = data
        self.steps = []
        self.original_columns = copy(data.columns)
        self.roles = self.data.roles
        self.columns = self.data.columns

        if outcomes:
            self.update_roles(outcomes, "outcome")
        if predictors:
            self.update_roles(predictors, "predictor")
        if groups:
            self.update_roles(groups, "group")
        if sequences:
            self.update_roles(sequences, "sequence")

    def add_roles(self, vars: Union[str, list[str]], new_role: str = "predictor") -> Recipe:
        """Adds an additional role for one or more columns of the Recipe's Ingredients.

        Args:
            vars: The column to receive additional roles.
            new_role: Defaults to predictor. The role to add to the column.

        See also:
            Ingredients.add_role()

        Returns:
            self
        """
        if isinstance(vars, str):
            vars = [vars]
        for v in vars:
            self.data.add_role(v, new_role)
        return self

    def update_roles(self, vars: Union[str, list[str]], new_role: str = "predictor", old_role: str = None) -> Recipe:
        """Adds a new role for one or more columns of the Recipe's Ingredients without roles
        or changes an existing role to a different one.

        Args:
            vars: The column to receive additional roles.
            new_role: Defaults to predictor'. The role to add or change to.
            old_role: Defaults to None. The role to be changed.

        See also:
            Ingredients.update_role()

        Returns:
            self
        """
        if isinstance(vars, str):
            vars = [vars]
        for v in vars:
            self.data.update_role(v, new_role, old_role)
        return self

    def add_step(self, step: Step) -> Recipe:
        """Adds a new step to the Recipe

        Args:
            step: a transformation step that should be applied to the Ingredients during prep() and bake()

        Returns:
            self
        """
        self.steps.append(step)
        return self

    def _check_data(self, data: Union[pl.DataFrame | pd.DataFrame, Ingredients]) -> Ingredients:
        if data is None:
            data = self.data
        elif isinstance(data, pl.DataFrame) or isinstance(data, pd.DataFrame):
            # this is only executed when prep or bake receive a DF that is different to the original data
            # don't check the roles here, because self.data can have more roles than data (post feature generation)
            data = Ingredients(data, roles=self.data.roles, check_roles=False)
        # if not data.columns.equals(self.data.columns):
        if not set(data.columns) == set(self.original_columns):
            raise ValueError(
                f"Columns of data argument differs from recipe data: "
                f"{[x for x in data.columns if x not in self.original_columns]}."
            )
        return data

    def _apply_group(self, data, step):
        if step.group:
            group_vars = select_groups(data)
            if len(group_vars) > 0:
                data.groupby(group_vars)
        return data

    def prep(
        self, data: Union[pl.DataFrame | pd.DataFrame, Ingredients] = None, refit: bool = False
    ) -> pl.DataFrame | pd.DataFrame:
        """Fits and transforms, in other words preps, the data.

        Args:
            data: Data to fit and transform. Defaults to None.
            refit: Defaults to False. Whether to refit data.

        Returns:
            Transformed data.
        """
        data = self._check_data(data)
        # Todo: check why the roles dissapear after copying
        data = copy(data)
        data = self._apply_fit_transform(data, refit)
        # return pl.DataFrame(data)
        return data.get_df()

    def bake(self, data: Union[pl.DataFrame | pd.DataFrame, Ingredients] = None) -> pl.DataFrame | pd.DataFrame:
        """Transforms, or bakes, the data if it has been prepped.

        Args:
            data: Data to transform. Defaults to None.

        Returns:
            Transformed data.
        """
        data = self._check_data(data)
        # original_data = deepcopy(data)
        data = self._apply_fit_transform(data)
        # return pl.DataFrame(data)
        return data.get_df()

    def _apply_fit_transform(self, data=None, refit=False):
        # applies transform or fit and transform (when refit or not trained yet)
        for step in self.steps:
            data = self._apply_group(data, step)
            if refit or not step.trained:
                data = step.fit_transform(data)
            else:
                data = step.transform(data)
        return data

    def __repr__(self):
        repr = "Recipe\n\n"

        # Print all existing roles and how many variables are assigned to each
        num_roles = Counter(chain.from_iterable(self.data.roles.values()))
        num_roles = pl.DataFrame({"role": [r for r in num_roles.keys()], "#variables": [n for n in num_roles.values()]})
        repr += "Inputs:\n\n" + num_roles.__repr__() + "\n\n"

        # Print all steps
        repr += "Operations:\n\n"
        for step in self.steps:
            repr += str(step) + "\n"

        return repr

    def get_backend(self):
        return self.data.get_backend()

    def cache(self):
        """Prepares the recipe for caching"""
        if self.data is not None:
            del self.data
        return self




import polars as pl
import numpy as np
from datetime import datetime, timedelta
import random
# Enable StringCache
pl.StringCache()




def generate_generic_datetime_data(stay_ids: list[int], max_rows: int, start_time: datetime, timestep: timedelta, columns: list[str], missingness: dict[str, float],
                                   categorical_columns=None) -> pl.DataFrame:
    """
    Generate a Polars DataFrame with the specified number of rows and schema for multiple stay IDs.

    Parameters:
    - stay_ids: The list of stay IDs to use.
    - max_rows: The maximum number of rows to generate.
    - start_time: The starting time for the time column.
    - timestep: The time interval between rows.
    - columns: A list of column names to generate data for.
    - missingness: A dictionary where keys are column names and values are the percentage of missing values.

    Returns:
    - A Polars DataFrame with the generated data.
    """
    # Distribute rows randomly among stay IDs
    rows_per_stay_id = np.random.multinomial(max_rows, [1/len(stay_ids)]*len(stay_ids))

    all_data = []

    for stay_id, num_rows in zip(stay_ids, rows_per_stay_id):
        # Generate the time column
        time_column = [start_time + i * timestep for i in range(num_rows)]

        # Initialize the data dictionary with the time and stay_id columns
        data = {
            "stay_id": [stay_id] * num_rows,
            "time": time_column
        }

        # Generate random data for other columns
        for column in columns:
            col_data = np.random.rand(num_rows)
            # Introduce missingness
            if column in missingness:
                missing_count = int(num_rows * missingness[column])
                missing_indices = np.random.choice(num_rows, missing_count, replace=False)
                col_data[missing_indices] = np.nan
            data[column] = col_data

        # Generate categorical data if specified
        if categorical_columns:
            for column, values in categorical_columns.items():
                col_data = random.choices(values, k=num_rows)
                # Introduce missingness
                if column in missingness:
                    missing_count = int(num_rows * missingness[column])
                    missing_indices = random.sample(range(num_rows), missing_count)
                    for idx in missing_indices:
                        col_data[idx] = None
                data[column] = col_data
                data[column] = pl.Series(data[column], dtype=pl.Categorical)
                # data[column] = pl.Series(data[column], dtype=pl.Enum)


        # Create the DataFrame for the current stay_id
        df = pl.DataFrame(data)
        all_data.append(df)

    # Concatenate all DataFrames
    final_df = pl.concat(all_data)

    return final_df

def generate_icu_data(max_rows: int, missingness_threshold: tuple[int,int]=(0.1,0.7), seed:int=None) -> pl.DataFrame:
    # Example usage
    if seed is not None:
        random.seed(seed)
    stay_ids = random.sample(range(1, max_rows*100), int(max_rows/24))
    start_time = datetime(2024, 1, 1)
    timestep = timedelta(hours=1)

    # Define the list of column names
    columns = ["alb", "alp", "alt", "ast", "be", "bicar", "bili", "bili_dir", "bnd", "bun", "ca", "cai", "ck",
               "ckmb", "cl", "crea", "crp", "dbp", "fgn", "fio2", "glu", "hgb", "hr", "inr_pt", "k", "lact", "lymph", "map",
               "mch", "mchc", "mcv", "methb", "mg", "na", "neut", "o2sat", "pco2", "ph", "phos", "plt", "po2", "ptt", "resp",
               "sbp", "temp", "tnt", "urine", "wbc"]
    categorical_columns = {"surgery":["Cardiac", "General", "Orthopedic", "Neurosurgery", "Vascular"],
                           "drainage":["Chest", "Abdominal", "Other"],
                           "ventilator":["Yes", "No"],
                           }
    # Define the missingness dictionary
    missingness = {item:random.uniform(missingness_threshold[0], missingness_threshold[1]) for item in columns}
    generated_df = generate_generic_datetime_data(stay_ids, max_rows, start_time, timestep, columns, missingness, categorical_columns)
    return generated_df







import os
import sys
sys.path.insert(0, os.path.abspath('../../'))  # Add your project root to Python path
# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'ReciPies'
copyright = '2025, Robin P. van de Water'
author = 'Robin P. van de Water'
release = '1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc']

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# Theme configuration
html_theme = 'sphinx_rtd_theme'

# Theme options
html_theme_options = {
    'canonical_url': '',
    'analytics_id': '',
    'logo_only': False,
    'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': False,
    'vcs_pageview_mode': '',
    'style_nav_header_background': 'white',
    # Toc options
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': 4,
    'includehidden': True,
    'titles_only': False
}

html_static_path = ['_static']

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',  # For Google/NumPy style docstrings
]

# Autodoc settings
autodoc_default_options = {
    'members': True,
    'member-order': 'bysource',
    'special-members': '__init__',
    'undoc-members': True,
    'exclude-members': '__weakref__'
}

# Generate autosummary even if no references
autosummary_generate = True


from abc import abstractmethod
from copy import deepcopy
from typing import Union, Dict

from pandas.core.groupby import DataFrameGroupBy
import polars as pl
from polars.dataframe.group_by import GroupBy
from scipy.sparse import isspmatrix

import pandas as pd
import numpy as np
from sklearn.preprocessing import StandardScaler

from pandas.api.types import is_timedelta64_dtype, is_datetime64_any_dtype

from src.recipies.ingredients import Ingredients
from enum import Enum
from src.recipies.selector import (
    Selector,
    all_predictors,
    all_numeric_predictors,
    select_groups,
    select_sequence,
)
from src.recipies.constants import Backend


class Step:
    """This class represents a step in a recipe.

    Steps are transformations to be executed on selected columns of a DataFrame.
    They fit a transformer to the selected columns and afterwards transform the data with the fitted transformer.

    Args:
        sel: Object that holds information about the selected columns.

    Attributes:
        columns: List with the names of the selected columns.
    """

    def __init__(self, sel: Selector = all_predictors(), supported_backends: list[Backend] = [Backend.POLARS, Backend.PANDAS]):
        self.sel = sel
        self.columns = []
        self._trained = False
        self._group = True
        self.supported_backends = supported_backends

    @property
    def trained(self) -> bool:
        return self._trained

    @property
    def group(self) -> bool:
        return self._group

    def fit(self, data: Ingredients):
        """This function fits the transformer to the data.

        Args:
            data: The DataFrame to fit to.
        """
        data = self._check_ingredients(data)
        self.columns = self.sel(data)
        self.do_fit(data)
        self._trained = True

    @abstractmethod
    def do_fit(self, data: Ingredients):
        pass

    def _check_ingredients(self, data: Union[Ingredients, GroupBy | DataFrameGroupBy]) -> Ingredients:
        """Check input for allowed types

        Args:
            data: input to the step

        Raises:
            ValueError: If a grouped pd.DataFrame is provided to a step that can't use groups.
            ValueError: If input are not (potentially grouped) Ingredients.

        Returns:
            Validated input
        """
        if self.supported_backends is not None and data.get_backend() not in self.supported_backends:
            raise ValueError(f"{data.get_backend()} not supported by this step.")
        if isinstance(data, GroupBy) or isinstance(data, DataFrameGroupBy):
            if not self._group:
                raise ValueError("Step does not accept grouped data.")
            # data = data.apply(lambda df: df)
        if not isinstance(data, Ingredients):
            raise ValueError(f"Expected Ingredients object, got {data.__class__}")
        return data

    def transform(self, data: Ingredients) -> Ingredients:
        """This function transforms the data with the fitted transformer.

        Args:
            data: The DataFrame to transform.

        Returns:
            The transformed DataFrame.
        """
        pass

    def fit_transform(self, data: Ingredients) -> Ingredients:
        self.fit(data)
        return self.transform(data)

    def __repr__(self) -> str:
        repr = self.desc + " for "

        if not self.trained:
            repr += str(self.sel)
        else:
            repr += str(self.columns) if len(self.columns) < 3 else str(self.columns[:2] + ["..."])  # FIXME: remove brackets
            repr += " [trained]"

        return repr


class StepImputeFill(Step):
    """For Pandas: uses pandas' internal `nafill` function to replace missing values.
    See `pandas.DataFrame.nafill` for a description of the arguments.
    """

    def __init__(self, sel=all_predictors(), value=None, strategy=None, limit=None):
        super().__init__(sel)
        self.desc = f"Impute with {strategy if strategy else value}"
        self.value = value
        self.strategy = strategy
        self.limit = limit

    def transform(self, data):
        new_data = self._check_ingredients(data)
        groups = select_groups(new_data)
        if data.get_backend() == Backend.POLARS:
            if len(groups) > 0:
                new_data.data = data.data.with_columns(
                    pl.col(self.columns).fill_null(self.value, strategy=self.strategy, limit=self.limit).over(groups)
                )
            else:
                new_data.data = data.data.with_columns(
                    pl.col(self.columns).fill_null(self.value, strategy=self.strategy, limit=self.limit)
                )
        else:
            # Pandas syntax
            func = None
            if self.strategy == "forward":
                func = pd.core.groupby.SeriesGroupBy.ffill
            elif self.strategy == "backward":
                func = pd.core.groupby.SeriesGroupBy.bfill
            elif self.strategy == "zero":
                self.value = 0
            elif self.value is None:
                raise ValueError(f"No valid strategy provided. Strategy was: {self.strategy}")

            if len(groups) > 0:
                df = new_data.groupby(groups)
            else:
                df = new_data.get_df()
            # [self.columns] = data.groupby(groups)[self.columns].fillna(self.value, method=self.strategy, limit=self.limit)
            new_df = new_data.get_df()
            if self.value is not None:
                # If value is set, fill with value
                if isinstance(df, GroupBy) or isinstance(df, DataFrameGroupBy):
                    # This type of imputation can only be done with ungrouped data
                    df = df.obj
                updated_columns = {col: df[col].fillna(self.value) for col in self.columns}
            else:
                # if func is None:
                #     # updated_columns = {col: df[col].fillna(self.value, method=self.strategy, limit=self.limit) for col in
                #     #                    self.columns}
                # else:
                updated_columns = {col: func(df[col]) for col in self.columns}

            # Use pd.concat to update the DataFrame in one go
            new_df = pd.concat([new_df.drop(columns=self.columns), pd.DataFrame(updated_columns)], axis=1)
            df = new_df
            new_data.set_df(df)
        return new_data


class StepImputeFastZeroFill(Step):
    """Quick variant of pandas' internal `nafill(value=0)` for grouped dataframes."""

    def __init__(self, sel=all_predictors()):
        super().__init__(sel, supported_backends=[Backend.PANDAS])
        self.desc = "Impute quickly with 0"

    def transform(self, data):
        new_data = self._check_ingredients(data)

        # Ignore grouping as grouping does not matter for zero fill.
        new_data[self.columns] = new_data[self.columns].fillna(0)

        return new_data


class StepImputeFastForwardFill(Step):
    """Quick variant of pandas' internal `nafill(method='ffill')` for grouped dataframes.

    Note: this variant does not allow for setting a limit.
    """

    def __init__(self, sel=all_predictors()):
        super().__init__(sel, supported_backends=[Backend.PANDAS])
        self.desc = "Impute with fast ffill"

    def transform(self, data):
        new_data = self._check_ingredients(data)

        # Use cumsum (which is optimised for grouped frames) to figure out which
        # values should be left at NaN, then ffill on the ungrouped dataframe. Adopted from:
        # https://stackoverflow.com/questions/36871783/fillna-forward-fill-on-a-large-dataframe-efficiently-with-groupby
        df = new_data.get_df()
        nofill = df.copy()
        nofill[self.columns] = pd.notnull(nofill[self.columns])
        nofill = nofill.groupby(select_groups(new_data))[self.columns].cumsum()

        df[self.columns] = df[self.columns].ffill()
        for col in self.columns:
            df.loc[nofill[col].to_numpy() == 0, col] = np.nan
        new_data.set_df(df)
        return new_data


# class StepImputeFastZeroFill(Step):
#     """Quick variant of pandas' internal `nafill(value=0)` for grouped dataframes."""
#
#     def __init__(self, sel=all_predictors()):
#         super().__init__(sel)
#         self.desc = "Impute quickly with 0"
#
#     def transform(self, data):
#         new_data = self._check_ingredients(data)
#         # Ignore grouping as grouping does not matter for zero fill.
#         new_data[self.columns] = new_data[self.columns].fillna(0)
#
#         return new_data
#
#
# class StepImputeFastForwardFill(Step):
#     """Quick variant of pandas' internal `nafill(method='ffill')` for grouped dataframes.
#
#     Note: this variant does not allow for setting a limit.
#     """
#
#     def __init__(self, sel=all_predictors()):
#         super().__init__(sel)
#         self.desc = "Impute with fast ffill"
#
#     def transform(self, data):
#         new_data = self._check_ingredients(data)
#
#         # Use cumsum (which is optimised for grouped frames) to figure out which
#         # values should be left at NaN, then ffill on the ungrouped dataframe. Adopted from:
#         # https://stackoverflow.com/questions/36871783/fillna-forward-fill-on-a-large-dataframe-efficiently-with-groupby
#         nofill = new_data.copy()
#         nofill[self.columns] = pd.notnull(nofill[self.columns])
#         nofill = nofill.groupby(data.keys).cumsum()
#
#         new_data[self.columns] = new_data[self.columns].ffill()
#         for col in self.columns:
#             new_data.loc[nofill[col].to_numpy() == 0, col] = np.nan
#
#         return new_data


class StepImputeModel(Step):
    """Uses a pretrained imputation model to impute missing values.
    Args:
        model: A function that takes a dataframe and the grouping columns as input and
            returns a dataframe with imputed values without the grouping column.
    """

    def __init__(self, sel=all_predictors(), model=None):
        super().__init__(sel)
        self.desc = "Impute with pretrained imputation model"
        self.model = model

    def transform(self, data):
        new_data = self._check_ingredients(data)
        if data.get_backend() == Backend.POLARS:
            new_data[self.columns] = self.model(new_data[self.columns + select_groups(new_data)], select_groups(new_data))
        return new_data


class Accumulator(Enum):
    MAX = "max"
    MIN = "min"
    MEAN = "mean"
    MEDIAN = "median"
    COUNT = "count"
    VAR = "var"
    FIRST = "first"
    LAST = "last"


class StepHistorical(Step):
    """This step generates columns with a historical accumulator provided by the user.

    Args:
        fun: Instance of the Accumulator enumerable that signifies which type of historical accumulation
            to use (default is MAX).
        suffix: Defaults to none. Set the name to have the step generate new columns with this suffix
            instead of the default suffix.
        role: Defaults to 'predictor'. In case new columns are added, set their role to role.
    """

    def __init__(
        self,
        sel: Selector = all_numeric_predictors(),
        fun: Accumulator = Accumulator.MAX,
        suffix: str = None,
        role: str = "predictor",
    ):
        super().__init__(sel)

        self.desc = f"Create historical {fun}"
        self.fun = fun
        if suffix is None:
            try:
                suffix = fun.value
            except Exception:
                raise TypeError(f"Expected Accumulator enum for function, got {self.fun.__class__}")
        self.suffix = suffix
        self.role = role

    def transform(self, data: Ingredients) -> Ingredients:
        """
        Raises:
            TypeError: If the function is not of type Accumulator
        """

        new_data = self._check_ingredients(data)
        self.suffix = "_" + self.suffix
        new_columns = [c + self.suffix for c in self.columns]

        selected = new_data.data
        selected_cols = pl.col(self.columns)
        id = select_groups(new_data)
        if data.get_backend() == Backend.POLARS:
            if self.fun is Accumulator.MAX:
                res = selected.with_columns(selected_cols.cum_max().over(id).name.suffix(self.suffix))
            elif self.fun is Accumulator.MIN:
                res = selected.with_columns(selected_cols.cum_min().over(id).name.suffix(self.suffix))
            elif self.fun is Accumulator.MEAN:
                res = selected.with_columns(
                    selected_cols.rolling_mean(window_size=selected.height, min_samples=0).over(id).name.suffix(self.suffix)
                )
            elif self.fun is Accumulator.MEDIAN:
                res = selected.with_columns(
                    selected_cols.rolling_median(window_size=selected.height, min_samples=0).over(id).name.suffix(self.suffix)
                )
            elif self.fun is Accumulator.COUNT:
                res = selected.with_columns(selected_cols.cum_count().over(id).name.suffix(self.suffix))
            elif self.fun is Accumulator.VAR:
                res = selected.with_columns(
                    selected_cols.rolling_var(window_size=selected.height, min_samples=0).over(id).name.suffix(self.suffix)
                )
            else:
                raise TypeError(f"Expected Accumulator enum for function, got {self.fun.__class__}")
            new_data.set_df(res)
        else:
            data = data.groupby(id)
            if self.fun is Accumulator.MAX:
                res = data[self.columns].cummax(skipna=True)
            elif self.fun is Accumulator.MIN:
                res = data[self.columns].cummin(skipna=True)
            elif self.fun is Accumulator.MEAN:
                # Reset index, as we get back a multi-index, and we want a simple rolling index
                res = data[self.columns].expanding().mean().reset_index(drop=True)
            elif self.fun is Accumulator.MEDIAN:
                res = data[self.columns].expanding().median().reset_index(drop=True)
            elif self.fun is Accumulator.COUNT:
                res = data[self.columns].expanding().count().reset_index(drop=True)
            elif self.fun is Accumulator.VAR:
                res = data[self.columns].expanding().var().reset_index(drop=True)
            else:
                raise TypeError(f"Expected Accumulator enum for function, got {self.fun.__class__}")
            # df = new_data.get_df()
            # df[new_columns] = res
            # new_data.set_df(df)
            new_data.set_df(new_data.get_df().assign(**{new_columns[i]: res.iloc[:, i] for i in range(len(new_columns))}))

        for nc in new_columns:
            new_data.update_role(nc, self.role)

        return new_data


class StepSklearn(Step):
    """This step takes a transformer from scikit-learn and makes it usable as a step in a recipe.

    Args:
        sklearn_transformer: Instance of scikit-learn transformer that implements fit() and transform().
        columnwise: Defaults to False. Set to True to fit and transform the DF column by column.
        in_place: Defaults to True. Set to False to have the step generate new columns
            instead of overwriting the existing ones.
        role (str, optional): Defaults to 'predictor'. Incase new columns are added, set their role to role.
    """

    def __init__(
        self,
        sklearn_transformer: object,
        sel: Selector = all_predictors(),
        columnwise: bool = False,
        in_place: bool = True,
        role: str = "predictor",
    ):
        super().__init__(sel)
        self.desc = f"Use sklearn transformer {sklearn_transformer.__class__.__name__}"
        self.sklearn_transformer = sklearn_transformer
        self.columnwise = columnwise
        self.in_place = in_place
        self.role = role
        self._group = False

    def do_fit(self, data: Ingredients) -> Ingredients:
        """
        Raises:
            ValueError: If the transformer expects a single column but gets multiple.
        """
        if self.columnwise:
            self._transformers = {
                # copy the transformer so we keep the distinct fit for each column and don't just refit
                col: deepcopy(self.sklearn_transformer.fit(data[col]))
                for col in self.columns
            }
        else:
            try:
                # print(data[self.columns])
                self.sklearn_transformer.fit(data[self.columns])
            except ValueError as e:
                if "should be a 1d array" in str(e) or "Multioutput target data is not supported" in str(e):
                    raise ValueError(
                        "The sklearn transformer expects a 1d array as input. " "Try running the step with columnwise=True."
                    )
                raise

    def transform(self, data: Ingredients) -> Ingredients:
        """
        Raises:
            TypeError: If the transformer returns a sparse matrix.
            ValueError: If the transformer returns an unexpected amount of columns.
        """
        new_data = self._check_ingredients(data)

        if self.columnwise:
            for col in self.columns:
                new_cols = self._transformers[col].transform(new_data[col])
                if self.in_place and new_cols.ndim == 2 and new_cols.shape[1] > 1:
                    raise ValueError(
                        "The sklearn transformer returned more than one column. Try running the step with in_place=False."
                    )
                col_names = (
                    col
                    if self.in_place
                    else [f"{self.sklearn_transformer.__class__.__name__}_{col}_{i + 1}" for i in range(new_cols.shape[1])]
                )
                if data.get_backend() == Backend.POLARS:
                    if isinstance(col_names, str):
                        col_names = [col_names]
                    updated_cols = pl.from_numpy(new_cols, schema=col_names)
                    new_data.data = new_data.data.with_columns(updated_cols)
                else:
                    df = new_data.get_df()
                    df[col_names] = new_cols
                    new_data.set_df(df)
        else:
            new_cols = self.sklearn_transformer.transform(new_data[self.columns])
            if isspmatrix(new_cols):
                raise TypeError(
                    "The sklearn transformer returns a sparse matrix, "
                    "but recipes expects a dense numpy representation. "
                    "Try setting sparse_output=False or similar in the transformer initialization."
                )

            col_names = (
                self.columns
                if self.in_place
                else (
                    [f"{self.sklearn_transformer.__class__.__name__}_{self.columns[i]}" for i in range(new_cols.shape[1])]
                    if new_cols.shape[1] == len(self.columns)
                    else [f"{self.sklearn_transformer.__class__.__name__}_{i + 1}" for i in range(new_cols.shape[1])]
                )
            )
            if new_cols.shape[1] != len(col_names):
                raise ValueError(
                    "The sklearn transformer returned a different amount of columns. Try running the step with in_place=False."
                )

            new_data[col_names] = new_cols

        # set role of new columns
        if not self.in_place:
            for col in col_names:
                new_data.update_role(col, self.role)

        return new_data


class StepResampling(Step):
    def __init__(
        self,
        new_resolution: str = "1h",
        accumulator_dict: Dict[Selector, Accumulator] = {all_predictors(): Accumulator.LAST},
        default_accumulator: Accumulator = Accumulator.LAST,
    ):
        """This class represents a resampling step in a recipe.

        Args:
            new_resolution: Resolution to resample to.
            accumulator_dict: Supply dictionary with individual accumulation methods for each Selector.
            default_accumulator: Accumulator to use for variables not supplied in dictionary.
        """
        super().__init__()
        self.new_resolution = new_resolution
        self.acc_dict = accumulator_dict
        self.default_accumulator = default_accumulator
        self._group = True

    def do_fit(self, data: Ingredients):
        self._trained = True

    def transform(self, data):
        new_data = self._check_ingredients(data)

        # Check for and save first sequence role
        if select_sequence(new_data) is not None:
            sequence_role = select_sequence(new_data)[0]
        else:
            raise AssertionError("Sequence role has not been assigned, resampling step not possible")
        sequence_datatype = new_data.dtypes[sequence_role]

        # if not (isinstance(pl.datatypes.TemporalType,sequence_datatype)): #or is_datetime64_any_dtype(sequence_datatype)):
        if data.get_backend() == Backend.POLARS and not (
            sequence_datatype.is_temporal()
        ):  # or is_datetime64_any_dtype(sequence_datatype)):
            raise ValueError(f"Expected Timedelta or Timestamp object, got {sequence_role(data).__class__}")
        if data.get_backend() == Backend.PANDAS and not (
            is_timedelta64_dtype(sequence_datatype) or is_datetime64_any_dtype(sequence_datatype)
        ):
            raise ValueError(f"Expected Timedelta or Timestamp object, got {sequence_role(data).__class__}")

        # Dictionary with the format column: str , accumulator:str is created
        col_acc_map = {}
        # Go through supplied Selector, Accumulator pairs
        for selector, accumulator in self.acc_dict.items():
            selected_columns = selector(new_data)
            # Add variables associated with selector with supplied accumulator
            col_acc_map.update({col: accumulator.value for col in selected_columns})

        # Add non-specified variables, if not a sequence role
        col_acc_map.update(
            {
                col: self.default_accumulator.value
                for col in set(new_data.columns).difference(col_acc_map.keys())
                if col not in select_sequence(new_data)
            }
        )
        # acc_col_map = dict((v, k) for k, v in col_acc_map.items())
        if data.get_backend() == Backend.POLARS:
            from collections import defaultdict

            acc_col_map = defaultdict(list)
            for k, v in col_acc_map.items():
                acc_col_map[v].append(k)
            if len(select_groups(new_data)) > 0:
                grouping_role = select_groups(new_data)[0]
                # Resampling with the functions defined in col_acc_map
                new_data.set_df(new_data.get_df().sort(grouping_role, sequence_role).set_sorted(sequence_role))
                new_data.set_df(
                    new_data.get_df()
                    .upsample(every=self.new_resolution, time_column=sequence_role, group_by=grouping_role)
                    .with_columns(pl.col(acc_col_map["last"]).fill_null(strategy="forward"))
                    .with_columns(pl.col(acc_col_map["mean"]).fill_null(strategy="mean"))
                    .with_columns(pl.col(acc_col_map["max"]).fill_null(strategy="max"))
                    .with_columns(pl.col(grouping_role).fill_null(strategy="forward"))
                )
            else:
                new_data.set_df(new_data.get_df().sort(sequence_role).set_sorted(sequence_role))
                new_data.set_df(
                    new_data.get_df()
                    .upsample(every=self.new_resolution, time_column=sequence_role)
                    .with_columns(pl.col(acc_col_map["last"]).fill_null(strategy="forward"))
                    .with_columns(pl.col(acc_col_map["mean"]).fill_null(strategy="mean"))
                    .with_columns(pl.col(acc_col_map["max"]).fill_null(strategy="max"))
                )
        else:
            # Resampling with the functions defined in col_acc_map
            if len(select_groups(new_data)) > 0:
                df = data.groupby(select_groups(data))
            else:
                df = data.get_df()
            new_data.set_df(df.resample(self.new_resolution, on=sequence_role).agg(col_acc_map))

            # Remove multi-index in case of grouped data
            if isinstance(data.get_df(), DataFrameGroupBy):
                new_data = new_data.set_df(new_data.get_df().droplevel(select_groups(data.get_df().obj)))

            # Remove sequence index, while keeping column
            # new_data = new_data.set_df(new_data.get_df().reset_index(drop=False))
        return new_data


class StepScale(StepSklearn):
    """Provides a wrapper for a scaling with StepSklearn.
    Note that because SKlearn transforms None (nulls) to NaN, we have to revert.

    Args:
       with_mean: Defaults to True. If True, center the data before scaling.
       with_std: Defaults to True. If True, scale the data to unit variance (or equivalently, unit standard deviation).
       in_place: Defaults to True. Set to False to have the step generate new columns instead of overwriting the existing ones.
       role (str, optional): Defaults to 'predictor'. Incase new columns are added, set their role to role.
    """

    def __init__(self, sel=all_numeric_predictors(), with_mean: bool = True, with_std: bool = True, *args, **kwargs):
        super().__init__(
            sklearn_transformer=StandardScaler(with_mean=with_mean, with_std=with_std), sel=sel, in_place=True, *args, **kwargs
        )
        self.desc = "Scale with StandardScaler"

    def transform(self, data: Ingredients) -> Ingredients:
        data = super().transform(data)
        # Revert null to nan conversion done by sklearn
        if data.get_backend() == Backend.POLARS:
            data.set_df(data.get_df().with_columns(pl.col(self.columns).fill_nan(None)))
        # else:
        #     data.set_df(data.get_df()[self.columns].fillna(value=None))
        return data


class StepFunction(Step):
    """Provides a wrapper for a simple transformation function, without fitting."""

    def __init__(self, sel: Selector, function):
        super().__init__(sel=sel)
        self.function = function
        self._trained = True

    def transform(self, data: Ingredients) -> Ingredients:
        new_data = self._check_ingredients(data)
        new_data = self.function(new_data)
        return new_data



import pytest

from src.recipies.ingredients import Ingredients
from src.recipies.constants import Backend
import polars as pl


def test_pl_init_role(example_df):
    Ingredients(example_df, roles={"y": "outcome"})
    assert True


def test_pd_init_role(example_df):
    Ingredients(example_df, roles={"y": "outcome"})
    assert True


def test_pl_init_role_wrong_type(example_df):
    with pytest.raises(TypeError) as e_info:
        Ingredients(example_df, roles=["outcome"])
    assert e_info.match(f"Expected dict object for roles, got {[].__class__}")


def test_pd_init_role_wrong_type(example_df):
    with pytest.raises(TypeError) as e_info:
        Ingredients(example_df, roles=["outcome"], backend=Backend.PANDAS)
    assert e_info.match(f"Expected dict object for roles, got {[].__class__}")


def test_pl_init_role_typo(example_df):
    with pytest.raises(ValueError) as e_info:
        Ingredients(example_df, roles={"z": "outcome"})
    assert e_info.match(r"^Roles contains variable names that are not in the data.")


def test_pd_init_role_typo(example_df):
    with pytest.raises(ValueError) as e_info:
        Ingredients(example_df, roles={"z": "outcome"}, backend=Backend.PANDAS)
    assert e_info.match(r"^Roles contains variable names that are not in the data.")


def test_pl_init_role_copy(example_df):
    roles = {"y": ["outcome"]}
    ing = Ingredients(example_df, roles=roles)
    roles["x1"] = ["predictor"]
    assert ing.roles == {"y": ["outcome"]}


def test_pd_init_role_copy(example_df):
    roles = {"y": ["outcome"]}
    ing = Ingredients(example_df, roles=roles, backend=Backend.PANDAS)
    roles["x1"] = ["predictor"]
    assert ing.roles == {"y": ["outcome"]}


def test_pl_init_role_noncopy(example_df):
    roles = {"y": ["outcome"]}
    ing = Ingredients(example_df, copy=False, roles=roles)
    roles["x1"] = ["predictor"]
    assert ing.roles == {"x1": ["predictor"], "y": ["outcome"]}


def test_pd_init_role_noncopy(example_df):
    roles = {"y": ["outcome"]}
    ing = Ingredients(example_df, copy=False, roles=roles, backend=Backend.PANDAS)
    roles["x1"] = ["predictor"]
    assert ing.roles == {"x1": ["predictor"], "y": ["outcome"]}


def test_pl_reinit_copy(example_df):
    ing = Ingredients(example_df, roles={"y": ["outcome"]})
    reing = Ingredients(ing)
    reing.add_role("y", "predictor")
    assert ing.roles != reing.roles


def test_pd_reinit_copy(example_df):
    ing = Ingredients(example_df, roles={"y": ["outcome"]}, backend=Backend.PANDAS)
    reing = Ingredients(ing)
    reing.add_role("y", "predictor")
    assert ing.roles != reing.roles


def test_pl_reinit_noncopy(example_df):
    ing = Ingredients(example_df, roles={"y": ["outcome"]})
    reing = Ingredients(ing, copy=False)
    reing.add_role("y", "predictor")
    assert ing.roles == reing.roles


def test_pd_reinit_noncopy(example_df):
    ing = Ingredients(example_df, roles={"y": ["outcome"]}, backend=Backend.PANDAS)
    reing = Ingredients(ing, copy=False)
    reing.add_role("y", "predictor")
    assert ing.roles == reing.roles


def test_add_role(example_ingredients):
    example_ingredients.update_role("y", "first role")
    example_ingredients.add_role("y", "another role")
    assert example_ingredients.roles["y"] == ["first role", "another role"]


def test_add_role_na(example_ingredients):
    with pytest.raises(RuntimeError) as e_info:
        example_ingredients.add_role("y", "first role")
    assert e_info.match("has no roles yet")


def test_update_role_na(example_ingredients):
    example_ingredients.update_role("y", "first role")
    assert example_ingredients.roles["y"] == ["first role"]


def test_update_role_na_but_old_role(example_ingredients):
    with pytest.raises(ValueError) as e_info:
        example_ingredients.update_role("y", "first role", "imaginary role")
    assert e_info.match("does not have a role yet")


def test_update_role_implicit(example_ingredients):
    example_ingredients.update_role("y", "first role")
    example_ingredients.update_role("y", "updated role")
    assert example_ingredients.roles["y"] == ["updated role"]


def test_update_role_implicit_multiple(example_ingredients):
    example_ingredients.update_role("y", "first role")
    example_ingredients.add_role("y", "second role")
    with pytest.raises(ValueError) as e_info:
        example_ingredients.update_role("y", "updated role")
    assert e_info.match("has more than one current role")


def test_update_role_explicit(example_ingredients):
    example_ingredients.update_role("y", "first role")
    example_ingredients.update_role("y", "updated role", "first role")
    assert example_ingredients.roles["y"] == ["updated role"]


def test_update_role_explicit_multiple(example_ingredients):
    example_ingredients.update_role("y", "first role")
    example_ingredients.add_role("y", "second role")
    example_ingredients.update_role("y", "updated role", "first role")
    assert example_ingredients.roles["y"] == ["second role", "updated role"]


def test_update_role_typo(example_ingredients):
    example_ingredients.update_role("y", "first role")
    with pytest.raises(ValueError) as e_info:
        example_ingredients.update_role("y", "updated role", "firs role")
    assert e_info.match("not among current roles")


def test_inferring_backend_ingredients(example_df):
    ing = Ingredients(example_df)
    if isinstance(example_df, pl.DataFrame):
        assert ing.get_backend() == Backend.POLARS
    else:
        assert ing.get_backend() == Backend.PANDAS


def test_explicit_backend_ingredients(example_df):
    ing = Ingredients(example_df, backend=Backend.PANDAS)
    assert ing.get_backend() == Backend.PANDAS
    ing = Ingredients(example_df, backend=Backend.POLARS)
    assert ing.get_backend() == Backend.POLARS



import argparse
from datetime import datetime, timedelta
import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import polars as pl
from sklearn.impute import MissingIndicator
from sklearn.preprocessing import KBinsDiscretizer, StandardScaler, MinMaxScaler, MaxAbsScaler, RobustScaler, PowerTransformer, \
    QuantileTransformer, SplineTransformer

from generate_data import generate_icu_data
from src.recipies.constants import Backend
from src.recipies import Recipe
from src.recipies.selector import all_numeric_predictors, all_predictors
from src.recipies.step import StepScale, StepSklearn, StepImputeFill, StepHistorical, Accumulator
import logging
import polars.selectors as cs
from memory_profiler import memory_usage
from tqdm import tqdm

def benchmark_dynamic_recipe(data, backend):
    if backend == Backend.POLARS:
        numeric_predictors = data.select(pl.exclude("stay_id", "time")).columns
    else:
        numeric_predictors = list(set(data.columns).difference(["stay_id", "time"]))
    dyn_rec = Recipe(data, [],numeric_predictors, "stay_id", "time", backend=backend)

    dyn_rec.add_step(StepScale(all_numeric_predictors(backend)))

    dyn_rec.add_step(StepSklearn(MissingIndicator(features="all"), all_predictors(), in_place=False))
    # dyn_rec.add_step(StepImputeFastForwardFill())

    dyn_rec.add_step(StepImputeFill(all_predictors(), strategy="forward"))
    # dyn_rec.add_step(StepImputeFastZeroFill())
    dyn_rec.add_step(StepImputeFill(all_predictors(), strategy="zero"))

    dyn_rec = dynamic_feature_generation(dyn_rec, backend)
    data = dyn_rec.bake()
    # print(data.head())


def benchmark_step(data, backend, step):
    timer = datetime.now()
    if backend == Backend.POLARS:
        numeric_predictors = data.select(pl.exclude("stay_id", "time")).columns
    else:
        numeric_predictors = list(set(data.columns).difference(["stay_id", "time"]))

    recipe = Recipe(data, [],numeric_predictors, "stay_id", "time", backend=backend)
    recipe.add_step(step)
    data = recipe.bake()
    time_passed = datetime.now() - timer
    return time_passed


def dynamic_feature_generation(data, backend):
    data.add_step(StepHistorical(sel=all_numeric_predictors(), fun=Accumulator.MIN, suffix="min_hist"))
    data.add_step(StepHistorical(sel=all_numeric_predictors(), fun=Accumulator.MAX, suffix="max_hist"))
    data.add_step(StepHistorical(sel=all_numeric_predictors(), fun=Accumulator.COUNT, suffix="count_hist"))
    data.add_step(StepHistorical(sel=all_numeric_predictors(), fun=Accumulator.MEAN, suffix="mean_hist"))
    return data


def benchmark_backend(backend, data_size, seed):
    # metrics = {}
    df_missing = generate_icu_data(data_size, seed=seed)
    df_complete = generate_icu_data(data_size, missingness_threshold=(0,0), seed=seed)
    if backend == Backend.PANDAS:
        backend_name = "Pandas"
        df_missing = df_missing.to_pandas()
        df_complete = df_complete.to_pandas()
    else:
        backend_name = "Polars"
    steps_missing = [
            StepImputeFill(all_predictors(), strategy="forward"),
            StepSklearn(MissingIndicator(features="all"), all_predictors(), in_place=False)]
    steps_complete = [
        # StepSklearn(
        #  OrdinalEncoder(),
        #  sel=has_type([str(pl.Categorical(ordering="physical")) if backend == backend.POLARS else "category"]),
        #  in_place=False,),
        # StepSklearn(
        #     OneHotEncoder(sparse_output=False),
        #     sel=has_type([str(pl.Categorical(ordering="physical")) if backend == backend.POLARS else "category"]),
        #     in_place=False,
        # ),
        StepSklearn(
         KBinsDiscretizer(n_bins=2, strategy="uniform", encode="ordinal"),
         sel=all_numeric_predictors(),
         in_place=False,
        ),
        StepHistorical(sel=all_numeric_predictors(), fun=Accumulator.MEAN, suffix="Mean"),
        StepHistorical(sel=all_numeric_predictors(), fun=Accumulator.MIN, suffix="Min"),
        StepHistorical(sel=all_numeric_predictors(), fun=Accumulator.MAX, suffix="Max"),
        StepHistorical(sel=all_numeric_predictors(), fun=Accumulator.COUNT, suffix="Count"),]

    for item in [StandardScaler(), MinMaxScaler(), MaxAbsScaler(), RobustScaler()]:
        steps_complete.append(StepSklearn(item, sel=all_numeric_predictors()))

    for transformer in [PowerTransformer(), QuantileTransformer(n_quantiles=10), SplineTransformer()]:
        steps_complete.append(StepSklearn(transformer, sel=all_numeric_predictors(), in_place=False))

    # benchmark_dynamic_recipe(df, backend)
    results = []

    def run_step_benchmark(backend, backend_name, data_size, results, step, df):
        metrics = {}
        metrics["data_size"] = data_size
        mem_usage, time_passed = memory_usage((benchmark_step, (df, backend, step)), retval=True, max_iterations=1, interval=0.5)
        metrics["backend"] = backend_name
        if isinstance(step, StepSklearn):
            metrics["step"] = str(step.sklearn_transformer.__class__.__name__)
        elif isinstance(step, StepHistorical):
            metrics["step"] = f"Historical{step.suffix}"
        else:
            metrics["step"] = str(step.__class__.__name__)
        metrics["time_passed"] = time_passed / timedelta(microseconds=1)
        metrics["memory_usage"] = max(mem_usage)
        results.append(metrics)

    for step in steps_complete:
        run_step_benchmark(backend, backend_name, data_size, results, step, df_complete)

    for step in steps_missing:
        run_step_benchmark(backend, backend_name, data_size, results, step, df_missing)
    return results

def parse_args():
    parser = argparse.ArgumentParser(description="Benchmarking script for recipies with ICU data generation and preprocessing.",
                                     usage="python perform_benchmark.py --data_sizes 10000 100000 --seeds 42 41")
    parser.add_argument('--data_sizes', type=int, nargs='+', default=[1000], help='List of data sizes to benchmark.')
    parser.add_argument('--seeds', type=int, nargs='+', default=[42, 41], help='Random seeds for data generation.')
    return parser.parse_args()

if __name__ == "__main__":
    args = parse_args()
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
    data_sizes = args.data_sizes
    seeds = args.seeds

    # Detect if output is being redirected to a file
    use_tqdm = sys.stdout.isatty() and sys.stderr.isatty()

    # Create filename once at the beginning
    csv_filename = f'results_datasizes_{data_sizes}_seeds_{seeds}_datetime_{datetime.now():%Y-%m-%d_%H-%M-%S}.csv'
    size_results = []
    with pl.StringCache():
        data_size_iter = tqdm(data_sizes, desc="Processing data sizes", unit="size") if use_tqdm else data_sizes
        for data_size in data_size_iter:
            logging.info(f"Starting with data size {data_size}")
            timer = datetime.now()

            # Collect results for this data size
            seed_iter = tqdm(seeds, desc=f"Seeds for size {data_size}", unit="seed", leave=False) if use_tqdm else seeds
            for seed in seed_iter:
                timer = datetime.now()
                logging.info(f"Starting with seed {seed}")
                polars = benchmark_backend(Backend.POLARS, data_size, seed=seed)
                pandas = benchmark_backend(Backend.PANDAS, data_size, seed=seed)
                polars = [pl.from_dict(item) for item in polars]
                pandas = [pl.from_dict(item) for item in pandas]
                polars = pl.concat(polars, how="vertical_relaxed")
                pandas = pl.concat(pandas, how="vertical_relaxed")
                size_results.extend([polars, pandas])
                logging.info(f"Time taken for seed {seed}: {datetime.now() - timer}")

            # Process results for this data size
            df_size = pl.concat(size_results, how="vertical_relaxed")
            df_size = df_size.group_by(["data_size", "step", "backend"]).agg([
                pl.col("time_passed").mean().alias("duration_mean"),
                pl.col("time_passed").std().alias("duration_std"),
                pl.col("memory_usage").mean().alias("memory_mean"),
                pl.col("memory_usage").std().alias("memory_std")
            ])

            columns = ["duration_mean", "duration_std", "memory_mean", "memory_std"]
            df_size = (df_size
                      .pivot(on="backend", values=columns, index=["data_size", "step"])
                      .with_columns(
                          speed_difference=(pl.col("duration_mean_Pandas") - pl.col("duration_mean_Polars")),
                          speedup=(pl.col("duration_mean_Pandas") / pl.col("duration_mean_Polars"))
                      ))
            df_size = df_size.with_columns(cs.numeric().round(1))
            df_size = df_size.sort(by=["data_size", "step"])

            df_size.write_csv(csv_filename)

            logging.info(f"Results for data size {data_size} written to {csv_filename}")
            print(f"Completed data size {data_size}")
# if __name__ == "__main__":
#     args = parse_args()
#     logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
#     data_sizes = args.data_sizes
#     seeds = args.seeds
#
#     df = None
#     df_list = []
#     for data_size in data_sizes:
#         logging.info(f"Starting with data size {data_size}")
#         timer = datetime.now()
#         for seed in seeds:
#             timer = datetime.now()
#             logging.info(f"Starting with seed {seed}")
#             combined = {}
#             polars = benchmark_backend(Backend.POLARS, data_size, seed=seed)
#             pandas = benchmark_backend(Backend.PANDAS, data_size, seed=seed)
#             polars = [pl.from_dict(item) for item in polars]
#             pandas = [pl.from_dict(item) for item in pandas]
#             polars = pl.concat(polars, how="vertical_relaxed")
#             pandas = pl.concat(pandas, how="vertical_relaxed")
#             if df is not None:
#                 df = pl.concat([df, polars,pandas], how="vertical_relaxed")
#             else:
#                 df = pl.concat([polars,pandas], how="vertical_relaxed")
#             logging.info(f"Time taken for data size {data_size}: {datetime.now() - timer}")
#             df_list.append(df)
#
#     df = pl.concat(df_list, how="vertical_relaxed")
#     df = df.group_by(["data_size", "step", "backend"]).agg([
#         pl.col("time_passed").mean().alias("duration_mean"),
#         pl.col("time_passed").std().alias("duration_std"),
#         pl.col("memory_usage").mean().alias("memory_mean"),
#         pl.col("memory_usage").std().alias("memory_std")
#     ])
#     columns = ["duration_mean", "duration_std", "memory_mean", "memory_std"]
#
#     df = (df#.group_by(["data_size", "step", "backend"])#.agg(pl.col("time_passed_mean"), pl.col("time_passed_std"))
#           .pivot(on="backend",values=columns, index=["data_size", "step"])
#           .with_columns(speed_difference = (pl.col("duration_mean_Pandas") - pl.col("duration_mean_Polars")),
#                         speedup = (pl.col("duration_mean_Pandas") / pl.col("duration_mean_Polars"))))
#     df = df.with_columns(cs.numeric().round(1))
#         #  value_name="time_passed"))
#     df = df.sort(by=["data_size", "step"])
#     df.write_csv(f'results_datasizes_{[data_sizes]}_seeds_{[seeds]}_datetime_{datetime.now():%Y-%m-%d_%H-%M-%S%z}.csv')
#     print(df)
        # combined = pl.concat([pl.from_dict(combined), pl.from_dict(polars, strict=False), pl.from_dict(pandas, strict=False)], how="vertical_relaxed")
        # prefix = "pandas"
        # pandas = {f"{prefix}_{key}": value for key, value in pandas.items()}
        # prefix = "polars"
        # polars = {f"{prefix}_{key}": value for key, value in polars.items()}
        # combined = {**polars, **pandas}

        # df = generate_icu_data(1000000)
        # timer = datetime.now()
        # benchmark_dynamic_recipe(df, Backend.POLARS)
        # polars_time = datetime.now() - timer
        # timer = datetime.now()
        # benchmark_dynamic_recipe(df, Backend.PANDAS)
        # pandas_time = datetime.now() - timer
        # print(f"Polars: time taken: {polars_time}")
        # print(f"Pandas time taken: {pandas_time}")
        # print(f"Speedup: {pandas_time/polars_time}")



from copy import deepcopy
import pandas as pd
import polars as pl
from typing import overload
from src.recipies.constants import Backend


class Ingredients:
    """Wrapper around either polars.DataFrame to store columns roles (e.g., predictor)
        Due to the workings of polars, we do not subclass pl.dataframe anymore,
        but instead store the dataframe as an attribute.
    Args:
        roles: roles of DataFrame columns as (list of) strings.
            Defaults to None.
        check_roles: If set to false, doesn't check whether the roles match existing columns.
            Defaults to True.

    See also: pandas.DataFrame

    Attributes:
        roles (dict): dictionary of column roles
    """

    _metadata = ["roles"]

    def __init__(
        self,
        data: pl.DataFrame | pd.DataFrame = None,
        copy: bool = None,
        roles: dict = None,
        check_roles: bool = True,
        backend: Backend = None,
    ):
        if backend is None:
            if isinstance(data, pl.DataFrame):
                self.backend = Backend.POLARS
            elif isinstance(data, pd.DataFrame):
                self.backend = Backend.PANDAS
            elif isinstance(data, Ingredients):
                self.backend = data.get_backend()
            else:
                raise ValueError("Backend not specified and could not be inferred from data.")
        else:
            self.backend = backend
        if isinstance(data, pd.DataFrame) or isinstance(data, pl.DataFrame):
            if self.backend == Backend.POLARS:
                if isinstance(data, pd.DataFrame):
                    self.data = pl.DataFrame(data)
                elif isinstance(data, pl.DataFrame):
                    self.data = data
                else:
                    raise TypeError(f"Expected DataFrame, got {data.__class__}")
            elif self.backend == Backend.PANDAS:
                if isinstance(data, pd.DataFrame):
                    self.data = data
                if isinstance(data, pl.DataFrame):
                    self.data = data.to_pandas()
            else:
                raise ValueError(f"Backend {self.backend} not supported.")
            self.schema = self.get_schema()
            self.dtypes = self.get_schema()

        if isinstance(data, Ingredients) and roles is None:
            if copy is None or copy is True:
                self.roles = deepcopy(data.roles)
            else:
                self.roles = data.roles
            self.data = data.data
            self.schema = data.schema
            self.dtypes = self.schema

        elif roles is None:
            self.roles = {}
        elif not isinstance(roles, dict):
            raise TypeError(f"Expected dict object for roles, got {roles.__class__}")
        elif check_roles and not all(set(k).issubset(set(self.data.columns)) for k, v in roles.items()):
            raise ValueError(
                f"Roles contains variable names that are not in the data {list(roles.values())} {self.data.columns}."
            )
        # Todo: do we want to allow ingredients without grouping columns?
        # elif check_roles and select_groups(self) == []:
        #     raise ValueError("Roles are given but no groups are found in the data.")
        else:
            if copy is None or copy is True:
                self.roles = deepcopy(roles)
            else:
                self.roles = roles

    @property
    def _constructor(self):
        return Ingredients

    @property
    def columns(self):
        return self.data.columns

    def to_df(self, output_format=None) -> pl.DataFrame:
        """Return the underlying DataFrame.


        Returns:
            Self as DataFrame.
        """
        if output_format == Backend.POLARS:
            if self.backend == Backend.POLARS:
                return self.data
            else:
                return pl.DataFrame(self.data)
        elif output_format == Backend.PANDAS:
            if self.backend == Backend.POLARS:
                return self.data.to_pandas()
            else:
                return self.data
        else:
            return self.data

    def _check_column(self, column):
        if not isinstance(column, str):
            raise ValueError(f"Expected string, got {column}")
        if column not in self.columns:
            raise ValueError(f"{column} does not exist in this Data object")

    def _check_role(self, new_role):
        if not isinstance(new_role, str):
            raise TypeError(f"new_role must be string, was {new_role.__class__}")

    def add_role(self, column: str, new_role: str):
        """Adds an additional role for a column that already has roles.

        Args:
            column: The column to receive additional roles.
            new_role: The role to add to the column.

        Raises:
            RuntimeError: If the column has no role yet.
        """
        self._check_column(column)
        self._check_role(new_role)
        if column not in self.roles.keys():
            raise RuntimeError(f"{column} has no roles yet, use update_role instead.")
        self.roles[column] += [new_role]

    def update_role(self, column: str, new_role: str, old_role: str = None):
        """Adds a new role for a column without roles or changes an existing role to a different one.

        Args:
            column: The column to update the roles of.
            new_role: The role to add or change to.
            old_role: Defaults to None. The role to be changed.

        Raises:
            ValueError: If old_role is given but column has no roles.
                If old_role is given but column has no role old_role.
                If no old_role is given but column has multiple roles already.
        """
        self._check_column(column)
        self._check_role(new_role)
        if old_role is not None:
            if column not in self.roles.keys():
                raise ValueError(
                    f"Attempted to update role of {column} from {old_role} to {new_role} "
                    f"but {column} does not have a role yet."
                )
            elif old_role not in self.roles[column]:
                raise ValueError(
                    f"Attempted to set role of {column} from {old_role} to {new_role} "
                    f"but {old_role} not among current roles: {self.roles[column]}."
                )
            self.roles[column].remove(old_role)
            self.roles[column].append(new_role)
        else:
            if column not in self.roles.keys() or len(self.roles[column]) == 1:
                self.roles[column] = [new_role]
            else:
                raise ValueError(
                    f"Attempted to update role of {column} to {new_role} but "
                    f"{column} has more than one current roles: {self.roles[column]}"
                )

    def select_dtypes(self, include=None):
        # if(isinstance(include,[str])):
        dtypes = self.get_str_dtypes()
        selected = [key for key, value in dtypes.items() if value in include]
        return selected

    def get_dtypes(self):
        dtypes = list(self.schema.values())
        return dtypes

    def get_str_dtypes(self):
        """ "
        Helper function for polar dataframes to return schema with dtypes as strings
        """
        dtypes = self.get_schema()
        return {key: str(value) for key, value in dtypes.items()}
        # return list(map(dtypes, cast()))

    def get_schema(self):
        if self.backend == Backend.POLARS:
            return self.data.schema
        else:
            return self.data.dtypes

    def get_df(self):
        return self.to_df()

    def set_df(self, df):
        self.data = df

    def groupby(self, by):
        if self.backend == Backend.POLARS:
            self.data.group_by(by)
        else:
            return self.data.groupby(by)

    def get_backend(self):
        return self.backend

    def __setitem__(self, idx, val):
        if self.backend == Backend.POLARS:
            self.data[idx] = val
        else:
            if isinstance(idx, tuple):
                rows, column = idx
                self.data.loc[rows, column] = val
            else:
                # Use assign for single column assignment to avoid fragmentation
                if isinstance(idx, str):
                    self.data = self.data.assign(**{idx: val})
                else:
                    # For non-string indices, use a more efficient approach
                    # that avoids the fragmentation warning
                    import warnings
                    with warnings.catch_warnings():
                        warnings.filterwarnings("ignore",
                                              message="DataFrame is highly fragmented",
                                              category=pd.errors.PerformanceWarning)
                        self.data[idx] = val

    @overload
    def __getitem__(self, list: list[str]) -> pl.DataFrame:
        return self.data[list]

    def __getitem__(self, idx: int) -> pl.Series:
        return self.data[idx]



import re
from src.recipies.ingredients import Ingredients
from typing import Union
from polars import DataType
from src.recipies.constants import Backend


class Selector:
    """Class responsible for selecting the variables affected by a recipe step

    Args:
        description: Text used to represent Selector when printed in summaries
        names: Column names to select. Defaults to None.
        roles: Column roles to select, see also Ingredients. Defaults to None.
        types: Column data types to select. Defaults to None.
        pattern: Regex pattern to search column names with. Defaults to None.
    """

    def __init__(
        self,
        description: str,
        names: Union[str, list[str]] = None,
        roles: Union[str, list[str]] = None,
        types: Union[str, list[str]] = None,
        pattern: re.Pattern = None,
    ):
        self.description = description
        self.set_names(names)
        self.set_roles(roles)
        self.set_types(types)
        self.set_pattern(pattern)

    def set_names(self, names: Union[str, list[str]]):
        """Set the column names to select with this Selector

        Args:
            names: column names to select
        """
        self.names = enlist_str(names)

    def set_roles(self, roles: Union[str, list[str]]):
        """Set the column roles to select with this Selector

        Args:
            roles: column roles to select, see also Ingredients
        """
        self.roles = enlist_str(roles)

    def set_types(self, roles: Union[str, list[str]]):
        """Set the column data types to select with this Selector

        Args:
            roles: column data types to select
        """
        self.types = enlist_str(roles)
        # self.types = enlist_dt(roles)

    def set_pattern(self, pattern: re.Pattern):
        """Set the pattern to search with this Selector

        Args:
            pattern: Regex pattern to search column names with.
        """
        self.pattern = pattern

    def __call__(self, ingr: Ingredients) -> list[str]:
        """Select variables from Ingredients

        Args:
            ingr: object from which to select the variables

        Raises:
            TypeError: when something other than an Ingredient object is passed

        Returns:
            Selected variables.
        """

        if not isinstance(ingr, Ingredients):
            raise TypeError(f"Expected Ingredients, got {ingr.__class__}")

        vars = list(ingr.columns)
        # Pandas
        # vars = ingr.columns.tolist()

        if self.roles is not None:
            # for v, r in ingr.roles.items():
            #     print(v, r)
            #     print(intersection(r, self.roles))
            sel_roles = [v for v, r in ingr.roles.items() if intersection(r, self.roles)]
            vars = intersection(vars, sel_roles)

        if self.types is not None:
            sel_types = list(ingr.select_dtypes(include=self.types))  # .columns.tolist()
            vars = intersection(vars, sel_types)

        if self.names is not None:
            vars = intersection(vars, self.names)

        if self.pattern is not None:
            vars = list(filter(self.pattern.search, vars))

        return vars

    def __repr__(self):
        return self.description


def enlist_dt(x: Union[DataType, list[DataType], None]) -> Union[list[DataType], None]:
    """Wrap a pl datatype in a list if it isn't a list yet

    Args:
        x: object to wrap.

    Raises:
        TypeError: If neither a datatype nor a list of datatypes is passed

    Returns:
        _description_
    """
    if isinstance(x, DataType):
        return [x]
    elif isinstance(x, list):
        if not all(isinstance(i, DataType) for i in x):
            raise TypeError("Only lists of datatypes are allowed.")
        return x
    elif x is None:
        return x
    else:
        raise TypeError(f"Expected a pl datatype, got {x.__class__}")


def enlist_str(x: Union[str, list[str], None]) -> Union[list[str], None]:
    """Wrap a str in a list if it isn't a list yet

    Args:
        x: object to wrap.

    Raises:
        TypeError: If neither a str nor a list of strings is passed

    Returns:
        _description_
    """
    if isinstance(x, str):
        return [x]
    elif isinstance(x, list):
        if not all(isinstance(i, str) for i in x):
            raise TypeError("Only lists of str are allowed.")
        return x
    elif x is None:
        return x
    else:
        raise TypeError(f"Expected str or list of str, got {x.__class__}")


def intersection(x: list, y: list) -> list:
    """Intersection of two lists

    Note:
        maintains the order of the first list
        does not deduplicate items (i.e., does not return a set)

    Args:
        x: first list
        y: second list

    Returns:
        Elements in `x` that are also in `y`.
    """
    if isinstance(x, str):
        x = [x]
    if isinstance(y, str):
        y = [y]
    return [i for i in x if i in y]


def all_of(names: Union[str, list[str]]) -> Selector:
    """Define selector for any columns with one of the given names

    Args:
        names: names to select

    Returns:
        Object representing the selection rule.
    """
    return Selector(description=str(names), names=names)


def regex_names(regex: str) -> Selector:
    """Define selector for any columns where the name matches the regex pattern

    Args:
        pattern: string to be transformed to regex pattern to search for

    Returns:
        Object representing the selection rule.
    """
    pattern = re.compile(regex)
    return Selector(description=f"regex: {regex}", pattern=pattern)


def starts_with(prefix: str) -> Selector:
    """Define selector for any columns where the name starts with the prefix

    Args:
        prefix: prefix to search for

    Returns:
        Object representing the selection rule.
    """
    return regex_names(f"^{prefix}")


def ends_with(suffix: str) -> Selector:
    """Define selector for any columns where the name ends with the suffix

    Args:
        prsuffixefix: suffix to search for

    Returns:
        Object representing the selection rule.
    """
    return regex_names(f"{suffix}$")


def contains(substring: str) -> Selector:
    """Define selector for any columns where the name contains the substring

    Args:
        substring: substring to search for

    Returns:
        Object representing the selection rule.
    """
    return regex_names(f"{substring}")


def has_role(roles: Union[str, list[str]]) -> Selector:
    """Define selector for any columns with one of the given roles

    Args:
        roles: roles to select

    Returns:
       Object representing the selection rule.
    """
    return Selector(description=f"roles: {roles}", roles=roles)


def has_type(types: Union[str, list[str]]) -> Selector:
    """Define selector for any columns with one of the given types

    Args:
        types: data types to select

    Note:
        Data types are selected based on string representation as returned by `df[[varname]].dtype.name`.

    Returns:
        Object representing the selection rule.
    """
    return Selector(description=f"types: {types}", types=types)


def all_predictors() -> Selector:
    """Define selector for all predictor columns

    Returns:
        Object representing the selection rule.
    """
    sel = has_role(["predictor"])
    sel.description = "all predictors"
    return sel


def all_numeric_predictors(backend=Backend.POLARS) -> Selector:
    """Define selector for all numerical predictor columns

    Returns:
        Object representing the selection rule.
    """
    sel = all_predictors()
    # if backend == Backend.POLARS:
    sel.set_types(["Int8", "Int16", "Int32", "Int64", "Float32", "Float64",
                   "int16", "int32", "int64", "float16", "float32", "float64"])
    # else:
    #     sel.set_types([])
    sel.description = "all numeric predictors"
    return sel


def all_outcomes() -> Selector:
    """Define selector for all outcome columns

    Returns:
        Object representing the selection rule.
    """
    sel = has_role(["outcome"])
    sel.description = "all outcomes"
    return sel


def all_groups() -> Selector:
    """Define selector for all grouping variables

    Returns:
        Object representing the selection rule.
    """
    return Selector(description="all grouping variables", roles=["group"])


def select_groups(ingr: Ingredients) -> list[str]:
    """Select any grouping columns

    Defines and directly applies Selector(roles=["group"])

    Returns:
        grouping columns
    """
    groups = all_groups()(ingr)
    return groups


def all_sequences() -> Selector:
    """Define selector for all grouping variables

    Returns:
        Object representing the selection rule.
    """
    return Selector(description="all sequence variables", roles=["sequence"])


def select_sequence(ingr: Ingredients) -> list[str]:
    """Select any sequence columns

    Defines and directly applies Selector(roles=["sequence"])

    Returns:
        Grouping columns.
    """
    return all_sequences()(ingr)
